/*
   lex.c -- PIKT common lex routines

   $Header: /home/cvs/ucgsb/pikt/src/lib/lex.c,v 1.1 2000/02/13 04:46:25 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2000 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#include "common.h"

/* return whether the leading parenthesis in the supplied string is a
   logic conditional parenthesis, as opposed to an arithmetic parenthesis;
   basically, we count the number of left versus right parentheses, until
   we reach certain turning points, where if the number of counted left
   parentheses exceeds the right count, we know that we are entering a
   logical condition
 */
int
condparen (char *s, int context)
{
	register int i;
	int lp, rp;
	register unsigned char c, nc;
	int done;

	done = FALSE;
	lp = rp = 0;
	i = 0;
	while ((c = *(s+i))) {
		switch(c) {
			case '(':
				lp++;
				break;
			case ')':
				rp++;
				if (lp == rp) {
					if (context == STRING)
						rp--;	/* signalling cond */
					done = TRUE;
				}
				break;
			case '<':
				nc = *(s+i+1);
				switch(nc) {
					case ' ':
					case '>':
					case '\n':
					case '\t':
						done = TRUE;
						break;
					default:
						break;
				}
				break;
			case '>':
				nc = *(s+i+1);
				switch(nc) {
					case ' ':
					case '\n':
					case '\t':
						done = TRUE;
						break;
					default:
						break;
				}
				break;
			case '-':	/* e.g., (-e $name) */
				nc = *(s+i+1);
				if (strchr("ezfdcbplS",nc) && (lp == 1))
					done = TRUE;
				break;
			default:
				break;
		}
		if (done) break;
		i++;
	}
	return lp > rp ? TRUE : FALSE;
}

/* when processing a statement like

      $substitute($line, "t", "X", 2) eq "this is a Xest"

   the lexer will eventually pattern match

      "t", "X", 2) eq "this is a Xest"

   this function isolates the next function string, i.e.,

      "t"

   returning its length, in this example, 3
 */
int
fnstr (char *s)
{
	register int i, j;
	register unsigned char c;
	register int q;
	int done;

	done = FALSE;
	q = 0;
	i = 0;
	while ((c = *(s+i))) {
		switch(c) {
			case '"':
				if (*(s+i+1) == '\0')
					done = TRUE;
				else
					q++;
				break;
			case '.':
			case ',':
			case ')':
				if (even(q)) {
					done = TRUE;
					j = 1;
					while ((c = *(s+i-j)) != '\"') {
						if (isspace(c))
							j++;
						else {
							done = FALSE;
							break;
						}
					}
					if (done)
						i -= j;
				}
				break;
			default:
				break;
		}
		if (done) break;
		i++;
	}
	return i;
}

/* given a string, break it down into its component parts--pure text strings,
   surrounded by single quotes, and variables or functions, joined together
   by the concatenation operator . (dot);

   for example, the input string

      "User $uname has uid $uid > the 60000 upper limit"

   when decomposed becomes

      'User ' . $uname . ' has uid ' . $uid . ' > the 60000 upper limit'

   the function returns the length of the output string

   the decomposition output string is pushed back into the lexer input
   stream for subsequent reprocessing
 */
int
decompose (signed char *istr, signed char *ostr)
{
	register signed char *cp;
	register int i;
	register signed char c;
	int isvar;
	register int escape;
	register int brkt, prn;
	int decomposed;

	decomposed = FALSE;
	cp = istr;
	isvar = FALSE;
	escape = FALSE;
	brkt = prn = 0;
	i = 0;
	while ((c = *cp)) {
		if ((c == '\\') && ! escape)  {
			escape = TRUE;
			cp++;
			continue;
		}
		if (isvar) {
			if ((c == LB) && ! prn && ! escape) {
				brkt++;
				*(ostr+i++) = c;
			} else if ((c == RB) && ! prn && brkt && ! escape) {
				brkt--;
				*(ostr+i++) = c;
			} else if ((c == LP) && ! brkt && ! escape) {
				prn++;
				*(ostr+i++) = c;
			} else if ((c == RP) && ! brkt && prn && ! escape) {
				prn--;
				*(ostr+i++) = c;
			} else if (isalnum((int)c) || ((brkt > 0) && ! prn) || ((prn >0) && ! brkt)) {
				*(ostr+i++) = c;
			} else {
				if (ISVAR(c)) {
					*(ostr+i) = '\0';
					strcat(ostr, " . ");
					i += 3;
					*(ostr+i++) = c;
				} else if ((c != '\"') || *(cp+1)) {
					*(ostr+i) = '\0';
					strcat(ostr, " . \'");
					i += 4;
					if (c == '\"') 
						*(ostr+i++) = DQTPROXY;
					else if (c == '\'') 
						*(ostr+i++) = SQTPROXY;
					else
						*(ostr+i++) = c;
					isvar = FALSE;
				}
			}
		} else { 
			if (ISVAR(c)) {
				if ((cp != istr+1) || ((cp == istr+1) && (c != '$'))) {
					*(ostr+i) = '\0';
					strcat(ostr, "\' . \'");
					i += 5;
				}
				i--;
				*(ostr+i++) = c;
				isvar = TRUE;
				decomposed = TRUE;
			} else if (c == '\"') {
				if ((cp == istr) || ! *(cp+1))	/* beginning or end of istr */
					*(ostr+i++) = '\'';
				else
					*(ostr+i++) = DQTPROXY;
			} else if (c == '\'') {
				*(ostr+i++) = SQTPROXY;
			} else {
				*(ostr+i++) = c;
			}
		}
		escape = FALSE;
		cp++;
	}
	*(ostr+i) = '\0';
	/* the following if is for the case where the istr is, e.g.,
	   "\$trim", and where regtok would (mistakenly) detect an
	   attempt to use a reserved function name as a var name;
	 */
	if (*(istr+1) == '\\') {
		if (strchr("$#%@",*(ostr+1)) && ! strncmp(istr+2, ostr+1, (size_t)(i-2))) {
			return 0;
		}
	}
	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "istr is %s, ostr is %s, decomposed is %d, i is %d", istr, ostr, decomposed, i);
	if (decomposed)
		return i-1;
	else
		return 1-i;
}
